import os
import glob

import pandas as pd
import numpy as np
import scipy.sparse as sparse

from lib5c.plotters.matrix import plot_matrix
from lib5c.util.system import check_outdir

from hiclite.util.filenames import format_outfile


def assemble_matrix(args, config, data):
    outfile = format_outfile(config, args.rep, args.chrom, 'raw')
    if os.path.exists(outfile):
        print('raw matrix already exists, skipping')
        data['raw'] = sparse.load_npz(outfile)
        data['current'] = data['raw']
        return

    chrom = args.chrom
    resolution = config.getint('assemble_params', 'resolution')
    chunksize = config.getint('assemble_params', 'chunksize')

    rows = []
    cols = []
    vals = []

    print('assembling raw matrix')
    for filename in glob.glob(format_outfile(
            config, args.rep, args.chrom, 'input')):
        print(' processing file %s' % filename)
        reader = pd.read_csv(
            filename, sep='\t', chunksize=chunksize, usecols=[1, 2, 4, 5],
            names=['chrom1', 'start1', 'chrom2', 'start2'])

        for idx, chunk in enumerate(reader):
            if idx % 10 == 0:
                print(' processing chunk number %s' % idx)

            chunk['start1'] = (chunk['start1'] / resolution).astype(int)
            chunk['start2'] = (chunk['start2'] / resolution).astype(int)

            chunk.drop(chunk[chunk['chrom1'] != chrom].index, inplace=True)
            chunk.drop(chunk[chunk['chrom2'] != chrom].index, inplace=True)

            rows.append(chunk['start1'])
            cols.append(chunk['start2'])
            vals.append(np.ones(len(chunk.index)))

    row = np.concatenate(rows)
    col = np.concatenate(cols)
    val = np.concatenate(vals)
    size = max(np.max(row), np.max(col)) + 1
    matrix = sparse.coo_matrix((val, (row, col)), shape=(size, size)).tocsr()

    print(' saving and plotting raw matrix')
    check_outdir(outfile)
    sparse.save_npz(outfile, matrix)
    if config.getboolean('matrix', 'plot'):
        plot_matrix(matrix, outfile=format_outfile(
            config, args.rep, args.chrom, 'raw', ext='png'))

    data['raw'] = matrix
    data['current'] = data['raw']
