import os
from collections import OrderedDict

import numpy as np
import pandas as pd

from lib5c.util.bed import parse_feature_from_string

from hiclite.util.clusters import save_clusters, make_loop_id_for_cluster, \
    find_clusters, load_clusters
from hiclite.util.filenames import format_outfile
from hiclite.util.sparse_matrix import SparseContactMatrix as SCM


def apply_thresholds(obs, pvalues, exp, min_obs=10.0, max_pvalue=0.001,
                     min_fc=1.5):
    """
    Applies specific thresholds to some ``np.ndarray``s, returning a boolean
    ``np.ndarray`` that has True at positions that pass all the thresholds.

    Parameters
    ----------
    obs, pvalues, exp : np.ndarray
        The arrays that the thresholds will be applied to. Should all have the
        same shape.
    min_obs, max_pvalue, min_fc : float
        The thresholds to apply.

    Returns
    -------
    np.ndarray
        A boolean array with True at positions that pass all the thresholds.
    """
    return np.where(
        (obs >= min_obs) &
        (pvalues <= max_pvalue) &
        ((obs / exp) >= min_fc),
        True, False)


def make_clusters(obs_scm, exp_scm, pvalues_scm, pvalue=0.001, min_obs=10.0,
                  min_fc=1.5, min_size=3, connectivity=1):
    sig_scm = SCM.apply(
        apply_thresholds,
        obs_scm, pvalues_scm, exp_scm,
        min_obs=min_obs,
        max_pvalue=pvalue,
        min_fc=min_fc
    )
    clusters = find_clusters(sig_scm, connectivity=connectivity)
    return [c for c in clusters if len(c) >= min_size]


def merge_clusters(small, big):
    all_small = {pixel for cluster in small for pixel in cluster}
    return small + [cluster for cluster in big
                    if not any([pixel in all_small for pixel in cluster])]


def process_clusters(args, config, data):
    clusters_outfile = format_outfile(config, args.rep, args.chrom, 'clusters')
    loops_outfile = format_outfile(config, args.rep, args.chrom, 'loops')
    if os.path.exists(clusters_outfile) and os.path.exists(loops_outfile):
        print('clusters and loops already exist, skipping')
        data['loops'] = pd.read_csv(loops_outfile, sep='\t', index_col=0)
        data['clusters'] = load_clusters(clusters_outfile)
        return

    obs_scm = data['obs']
    exp_scm = data['exp']
    pvalues_scm = data['current']

    print('making clusters')
    print(' thresholding')
    max_pvalue = config.getfloat('clusters_params', 'max_pvalue')
    min_pvalue = config.getfloat('clusters_params', 'min_pvalue') \
        if config.has_option('clusters_params', 'min_pvalue') else max_pvalue

    clusters = reduce(
        merge_clusters,
        (make_clusters(
            obs_scm, exp_scm, pvalues_scm, pvalue=pvalue,
            min_obs=config.getfloat('clusters_params', 'min_obs'),
            min_fc=config.getfloat('clusters_params', 'min_fc'),
            min_size=config.getint('clusters_params', 'min_size'),
            connectivity=config.getint('clusters_params', 'connectivity')
         )
         for pvalue in reversed(
            10 ** np.arange(np.log10(max_pvalue),
                            np.log10(min_pvalue)-1, step=-1)))
    )

    print(' distance filtering')
    min_dist = config.getint('clusters_params', 'min_dist')
    clusters = [c for c in clusters if min([j-i for i, j in c]) >= min_dist]

    print(' saving clusters')
    save_clusters(clusters, clusters_outfile)

    print(' compiling loop information')
    loops = []
    for cluster in clusters:
        cluster = list(cluster)
        peak_pixel = cluster[np.argmin(pvalues_scm[tuple(zip(*cluster))])]
        loop_id = make_loop_id_for_cluster(
            cluster, args.chrom, config.getint('assemble_params', 'resolution'))
        us_anchor, ds_anchor = map(parse_feature_from_string,
                                   loop_id.split('_'))
        loops.append(OrderedDict([
            ('loop_id', loop_id),
            ('obs', float(obs_scm[peak_pixel])),
            ('exp', float(exp_scm[peak_pixel])),
            ('pvalue', float(pvalues_scm[peak_pixel])),
            ('cluster_size', len(cluster)),
            ('us_chrom', us_anchor['chrom']),
            ('us_start', us_anchor['start']),
            ('us_end', us_anchor['end']),
            ('ds_chrom', ds_anchor['chrom']),
            ('ds_start', ds_anchor['start']),
            ('ds_end', ds_anchor['end']),
            ('peak_pixel', make_loop_id_for_cluster(
                [peak_pixel], args.chrom,
                config.getint('assemble_params', 'resolution')))
        ]))
    if loops:
        df = pd.DataFrame(loops)
        df.set_index('loop_id', inplace=True)
        df.sort_values(by='us_start', inplace=True)

        print(' saving loop information')
        df.to_csv(loops_outfile, sep='\t')

        data['loops'] = df
    else:
        print(' no loops found, writing header line only')
        with open(loops_outfile, 'w') as handle:
            handle.write('loop_id\tobs\texp\tpvalue\tcluster_size\tus_chrom\t'
                         'us_start\tus_end\tds_chrom\tds_start\tds_end\t'
                         'peak_pixel\n')

        data['loops'] = None
    data['clusters'] = clusters
