import os

import numpy as np
import scipy.sparse as sparse

from lib5c.plotters.matrix import plot_matrix
from lib5c.util.system import check_outdir

from hiclite.util.filenames import format_outfile
from hiclite.util.sparse_matrix import SparseContactMatrix as SCM
from hiclite.util.outliers import remove_high_spatial_outliers
from lib5c.contrib.pybigwig.bigwig import bigwig_avail


def filter_sparse_rows_count(matrix, min_nnz=25, k=300):
    """
    Wipes the sparse bins (both rows and columns) of a matrix, where "sparse"
    means that the bin has less than ``min_nnz`` nonzero interactions with both
    the ``k`` nearest upstream and downstream bins.

    This function wipes with zeros and will eliminate the wiped positions if
    ``matrix`` is a CSR format sparse matrix.

    Parameters
    ----------
    matrix : np.ndarray, scipy.sparse.csr_matrix, or SparseContactMatrix
        The matrix to wipe sparse bins from.
    min_nnz : int
        The minimum number of nonzero contacts a bin has to make in either
        direction (upstream or downstream) to escape being wiped.
    k : int
        How many bins upstream or downstream to look when counting the number
        of nonzero contacts a bin makes in a given direction.

    Returns
    -------
    np.ndarray, scipy.sparse.csr_matrix, or SparseContactMatrix
        The filtered matrix.
    """
    matrix = matrix.copy()
    if min_nnz == 0 or k == 0:
        return matrix
    if isinstance(matrix, np.ndarray):
        # matrix already supports row/col indexing
        # scm can be created directly from matrix
        scm = SCM.from_ndarray(matrix, max_range=k)
    else:
        if SCM.is_scm(matrix):
            scm = matrix
        elif isinstance(matrix, sparse.spmatrix):
            if not isinstance(matrix, sparse.csr_matrix):
                # incoming matrix doesn't support row/col indexing, convert it
                matrix = matrix.tocsr()
            # create SCM directly from sparse matrix
            scm = SCM(matrix, max_range=k)
        else:
            raise ValueError('invalid input type')
    scm.symmetrize()
    us_rows = np.where((scm.offsets >= -k) & (scm.offsets < 0))[0]
    ds_rows = np.where((scm.offsets <= k) & (scm.offsets > 0))[0]
    deleted_indices = (np.sum(scm.data[us_rows, :] > 0, axis=0) < min_nnz) & \
                      (np.sum(scm.data[ds_rows, :] > 0, axis=0) < min_nnz)
    if isinstance(matrix, sparse.csr_matrix):
        keep_indices_csr = sparse.diags([~deleted_indices], [0], dtype=int)
        matrix = keep_indices_csr.dot(matrix)
        matrix = matrix.dot(keep_indices_csr)
    else:
        deleted_indices = np.where(deleted_indices)[0]
        matrix[:, deleted_indices] = 0
        matrix[deleted_indices, :] = 0
    return matrix


def filter_sparse_zones(matrix, zone_size=25):
    matrix = matrix.copy()
    if zone_size == 0:
        return matrix
    zero = matrix.diagonal() == 0  # bool vector, True where diag count is zero
    zero_idx = np.where(zero)[0]  # int vector of where the zeros are
    zero_idx_diff = np.diff(zero_idx)  # distances between consecutive zeros
    deleted_indices = np.zeros_like(zero)  # tracks what we choose to delete
    for i in range(len(zero_idx_diff)):
        if zero_idx_diff[i] < zone_size:
            deleted_indices[zero_idx[i]:zero_idx[i+1]] = True
    keep_indices_csr = sparse.diags([~deleted_indices], [0], dtype=int)
    matrix = keep_indices_csr.dot(matrix)
    matrix = matrix.dot(keep_indices_csr)
    return matrix


def filter_matrix(args, config, data):
    outfile = format_outfile(config, args.rep, args.chrom, 'filtered')
    if os.path.exists(outfile):
        print('filtered matrix already exists, skipping')
        data['filtered'] = SCM.load(outfile) if SCM.is_scm(outfile)\
            else sparse.load_npz(outfile)
        data['current'] = data['filtered']
        return

    raw = data['current']

    print('filtering matrix')
    if config.getfloat('filter_params', 'map_threshold') and not bigwig_avail():
        raise ImportError('importing pyBigWig failed, is it installed?')
    if config.getfloat('filter_params', 'map_threshold'):
        from hiclite.util.mappability import make_mappability_table, \
            make_mappability_vector, make_bin_mappability_vector,\
            filter_low_mappability_bins
        vector_file = format_outfile(config, args.rep, args.chrom,
                                     'mappability')
        check_outdir(vector_file)
        if not os.path.exists(vector_file):
            if config.getint('filter_params', 'map_window'):
                vector = make_bin_mappability_vector(
                    config.get('filter_params', 'mappability_bigwig'),
                    args.chrom,
                    resolution=config.getint('assemble_params', 'resolution'),
                    window_width=config.getint('filter_params', 'map_window')
                )
            else:
                vector = make_mappability_vector(
                    make_mappability_table(
                        config.get('filter_params', 'fragment_bedfile'),
                        config.get('filter_params', 'mappability_bigwig'),
                        args.chrom
                    ),
                    resolution=config.getint('assemble_params', 'resolution')
                )
            np.savetxt(vector_file, vector)
        vector = np.loadtxt(vector_file)
        if len(vector) > raw.shape[0]:
            vector = vector[:raw.shape[0]]
        filtered = filter_low_mappability_bins(
            raw,
            vector,
            threshold=config.getfloat('filter_params', 'map_threshold')
        )
    else:
        filtered = raw
    filtered = filter_sparse_zones(
        filter_sparse_rows_count(
            filtered,
            min_nnz=config.getint('filter_params', 'sparse_count'),
            k=config.getint('filter_params', 'sparse_range'),
        ),
        zone_size=config.getint('filter_params', 'zone_size')
    )
    if config.getfloat('filter_params', 'outlier_threshold'):
        filtered = remove_high_spatial_outliers(
            filtered,
            max_range=config.getint('obs_params', 'max_range'),
            size=config.getint('filter_params', 'outlier_size'),
            fold_threshold=config.getfloat('filter_params', 'outlier_threshold')
        )

    print(' saving and plotting filtered matrix')
    if isinstance(raw, SCM):
        filtered.save(outfile)
    else:
        check_outdir(outfile)
        sparse.save_npz(outfile, filtered)
    if config.getboolean('matrix', 'plot'):
        plot_matrix(filtered, outfile=format_outfile(
            config, args.rep, args.chrom, 'filtered', ext='png'))

    data['filtered'] = filtered
    data['current'] = data['filtered']
