import os

import numpy as np
import scipy.stats as stats

from lib5c.plotters.matrix import plot_matrix
from lib5c.util.distributions import call_pvalues

from hiclite.util.sparse_matrix import SparseContactMatrix as SCM
from hiclite.util.filenames import format_outfile


def process_pvalues(args, config, data):
    outfile = format_outfile(config, args.rep, args.chrom, 'pvalues')
    if os.path.exists(outfile):
        print('pvalues already exist, skipping')
        data['pvalues'] = SCM.load(outfile)
        data['current'] = data['pvalues']
        return

    obs_scm = data['obs']
    exp_scm = data['exp']
    var_scm = data['current']

    print('making pvalues')
    # check deconv
    if config.getboolean('var_params', 'deconvolute'):
        print(' deconvoluting')
        bias = data['bias']
        obs_scm = obs_scm.deconvolute(bias)
        exp_scm = exp_scm.deconvolute(bias)

    # check clamp
    min_exp = config.getfloat('var_params', 'min_exp')
    if min_exp:
        print(' clamping expected values')
        exp_scm = SCM.apply(lambda x, y: np.where(x < y, y, x),
                            exp_scm, min_exp)

    # decide dist_gen and log
    if config.get('pvalues_params', 'model') == 'loglogistic':
        dist_gen = stats.logistic
        log = True
    elif config.get('pvalues_params', 'model') == 'lognorm':
        dist_gen = stats.norm
        log = True
    else:
        dist_gen = getattr(stats, config.get('pvalues_params', 'model'))
        log = False

    # call pvalues
    pvalues_scm = SCM.apply(call_pvalues, np.log(obs_scm) if log else obs_scm,
                            exp_scm, var_scm, dist_gen, log=log)

    print(' saving and plotting pvalues')
    pvalues_scm.save(outfile)
    if config.getboolean('matrix', 'plot'):
        plot_matrix(pvalues_scm, outfile=format_outfile(
            config, args.rep, args.chrom, 'pvalues', ext='png'))

    data['pvalues'] = pvalues_scm
    data['current'] = data['pvalues']
