import os
import subprocess
import shutil

import numpy as np
import scipy.stats as stats

from lib5c.plotters.fits import plot_fit
from lib5c.plotters.matrix import plot_matrix
from lib5c.util.system import check_outdir

from hiclite.util.sparse_matrix import SparseContactMatrix as SCM
from hiclite.util.filenames import format_outfile
from hiclite.util.mvr import fit_mvr
from hiclite.util.cross_rep_variance import fit_cross_rep_variance, disp_to_var
from hiclite.util.cmd import make_process_cmd


def process_var(args, config, data):
    outfile = format_outfile(config, args.rep, args.chrom, 'var')
    if os.path.exists(outfile):
        print('var already exists, skipping')
        data['var'] = SCM.load(outfile)
        data['current'] = data['var']
        return

    obs_scm = data['obs']
    exp_scm = data['current']

    print('making var')

    # short-circuit for poisson
    if config.get('var_params', 'method') == 'poisson':
        check_outdir(outfile)
        shutil.copyfile(format_outfile(config, args.rep, args.chrom, 'exp'),
                        outfile)
        data['var'] = exp_scm
        data['current'] = data['var']
        return

    if config.getboolean('var_params', 'deconvolute'):
        if config.get('var_params', 'method') == 'crv':
            raise ValueError('deconvolution not yet supported for cross-rep '
                             'variance method')
        print(' deconvoluting')
        bias = data['bias']
        obs_scm = obs_scm.deconvolute(bias)
        if config.get('var_params', 'method') != 'deseq2':
            exp_scm = exp_scm.deconvolute(bias)
    obs = obs_scm.data
    exp = exp_scm.data

    def load_other_reps():
        print(' loading other replicates')
        _obs_scms = []
        for rep in zip(*config.items('reps'))[0]:
            if rep == args.rep:
                _obs_scms.append(obs_scm)
                continue
            filename = format_outfile(config, rep, args.chrom, 'obs')
            if os.path.exists(filename):
                _scm = SCM.load(filename)
            else:
                print(' making obs for rep %s' % rep)
                cmd = make_process_cmd(args, config, stop_at='obs',
                                       override_rep=rep)
                print(cmd)
                subprocess.call(cmd, shell=True)
                _scm = SCM.load(filename)
            if config.getboolean('var_params', 'deconvolute'):
                print(' deconvoluting rep %s' % rep)
                _scm.deconvolute(np.loadtxt(format_outfile(
                    config, rep, args.chrom, 'bias')))
            _obs_scms.append(_scm)
        return _obs_scms

    if config.get('var_params', 'method') == 'obs_over_exp':
        print(' computing obs/exp')
        obs_over_exp = obs / exp
        if config.getboolean('matrix', 'plot'):
            obs_over_exp_scm = SCM(
                (obs_over_exp, obs_scm.offsets), shape=obs_scm.shape)
            plot_matrix(obs_over_exp_scm, log=True, outfile=format_outfile(
                    config, args.rep, args.chrom, 'obs_over_exp', ext='png'))

        print(' fitting logistic')
        finite_indices = np.isfinite(obs_over_exp) & (obs_over_exp > 0)
        data = np.log(obs_over_exp[finite_indices])
        fitted_dist = stats.logistic(*stats.logistic.fit(data))
        plot_fit(data, fitted_dist, outfile=format_outfile(
            config, args.rep, args.chrom, 'fitting'))
        disp = fitted_dist.stats('v')
        var = disp_to_var(disp, exp)

    elif config.get('var_params', 'method') == 'mvr':
        print(' fitting mvr')
        var = fit_mvr(obs, exp,
                      min_obs=config.getfloat('var_params', 'min_obs'),
                      min_exp=config.getfloat('var_params', 'min_exp'))

    elif config.get('var_params', 'method') == 'crv':
        obs_scms = load_other_reps()
        print(' computing cross-replicate variance')
        var = fit_cross_rep_variance(
            [scm.data for scm in obs_scms], exp,
            min_obs=config.getfloat('var_params', 'min_obs'),
            min_exp=config.getfloat('var_params', 'min_exp'),
            min_dist=config.getint('var_params', 'min_dist'))

    elif config.get('var_params', 'method') == 'deseq2':
        try:
            from hiclite.contrib.deseq2 import compute_dispersions,\
                compute_variance
        except ImportError, e:
            print('rpy2 not installed, install it and try again')
            raise e
        obs_scms = load_other_reps()
        print(' computing dispersions')
        plot_outfile = format_outfile(
            config, args.rep, args.chrom, 'disp', ext='png')
        obs_arrays = [scm.data for scm in obs_scms]
        disp = compute_dispersions(
            obs_arrays,
            min_obs=config.getfloat('var_params', 'min_obs'),
            plot_outfile=plot_outfile)
        print(' computing variance')
        var = compute_variance(disp, obs_arrays)

    else:
        raise ValueError('method %s not recognized' %
                         config.get('var_params', 'method'))

    print(' saving and plotting var')
    var_scm = SCM((var, obs_scm.offsets), shape=obs_scm.shape)
    var_scm.save(outfile)
    if config.getboolean('matrix', 'plot'):
        plot_matrix(var_scm, outfile=format_outfile(
            config, args.rep, args.chrom, 'var', ext='png'))

    data['var'] = var_scm
    data['current'] = data['var']
