import ConfigParser
import subprocess
import os
import sys

import numpy as np
import pandas as pd
import scipy.sparse as sparse

from hiclite.util.filenames import format_outfile
from hiclite.util.sparse_matrix import SparseContactMatrix as SCM
from hiclite.util.cmd import make_process_cmd, make_collect_cmd
from hiclite.tools.config import drop_config
from hiclite.toolbox import subcommand, argument

from hiclite.steps.assemble import assemble_matrix
from hiclite.steps.filter import filter_matrix
from hiclite.steps.balance import balance_matrix
from hiclite.steps.obs import process_obs
from hiclite.steps.exp import process_exp
from hiclite.steps.var import process_var
from hiclite.steps.pvalues import process_pvalues
from hiclite.steps.qvalues import process_qvalues
from hiclite.steps.cluster import process_clusters


STEP_TO_FUNCTION = {
    'assemble': assemble_matrix,
    'filter': filter_matrix,
    'balance': balance_matrix,
    'obs': process_obs,
    'exp': process_exp,
    'var': process_var,
    'pvalues': process_pvalues,
    'clusters': process_clusters
}


def process(args, config):
    data = {}
    steps = zip(*config.items('steps'))[0]
    for step in steps:
        if step == 'collect':
            continue
        if step == 'qvalues':
            # presumably the previous step was pvalues, and so data['current']
            # points to pvalues; we want to overwrite this with the qvalues so
            # that subsequent steps (e.g., clustering) use the qvalues instead
            # of the pvalues
            # qvalues can't be computed within the standard step funciton API,
            # so they don't have an entry in STEP_TO_FUNCTION; instead, we
            # assume that if you've gotten this far you've passed the check in
            # main() that confirms that the qvalues output file already exists;
            # instead of computing qvalues by calling a function in
            # STEP_TO_FUNCTION and assuming that that function will update
            # data['current'] accordingly, we just assume the file already
            # exists, and then the qvalues step reduces to just loading that
            # file into data['current']
            data['qvalues'] = SCM.load(
                format_outfile(config, args.rep, args.chrom, 'qvalues'))
            data['current'] = data['qvalues']
            continue
        if 'current' not in data and step != 'assemble':
            infile = format_outfile(config, args.rep, args.chrom, 'input')
            if SCM.is_scm(infile):
                data['current'] = SCM.load(infile)
            else:
                data['current'] = sparse.load_npz(infile)
        if step in ['var', 'pvalues'] and 'bias' not in data and \
                config.getboolean('var_params', 'deconvolute'):
            data['bias'] = np.loadtxt(format_outfile(
                config, args.rep, args.chrom, 'bias'))
        STEP_TO_FUNCTION[step](args, config, data)
        if args.stop_at is not None and step == args.stop_at:
            break


def collect(args, config):
    df = pd.concat([
        pd.read_csv(format_outfile(config, args.rep, chrom, 'loops', ext='tsv'),
                    sep='\t', index_col=0)
        for chrom, _ in config.items('chroms')])
    df.to_csv(format_outfile(config, args.rep, 'all', 'loops', ext='tsv'),
              sep='\t')


@subcommand('pipeline', 'run the pipeline', [
    argument('action', choices=['process', 'collect', 'full', 'check']),
    argument('rep'),
    argument('--chrom', default='all'),
    argument('--stop_at'),
    argument('--config', default='config.cfg')
])
def main(args):
    if not os.path.isfile(args.config):
        print('config file not found, dropping default config')
        drop_config()
        return

    print('parsing config')
    config = ConfigParser.RawConfigParser(allow_no_value=True)
    config.optionxform = str
    config.read(args.config)

    if args.action == 'process':
        chroms = zip(*config.items('chroms'))[0]
        if 'qvalues' in zip(*config.items('steps'))[0] and args.stop_at is None:
            if not all(map(lambda x: os.path.exists(format_outfile(
                    config, args.rep, x, 'pvalues')), chroms)):
                print('qvalues in steps, but pvalues not available for all '
                      'chroms: running pipeline over all chroms with '
                      '`--stop_at pvalues` - please rerun your command after '
                      'this completes')
                for chrom in chroms:
                    cmd = make_process_cmd(
                        args, config, chrom=chrom, stop_at='pvalues')
                    print(cmd)
                    subprocess.call(cmd, shell=True)
                sys.exit(1)
            else:
                if not all(map(lambda x: os.path.exists(format_outfile(
                        config, args.rep, x, 'qvalues')), chroms)):
                    process_qvalues(args, config)
        if args.chrom == 'all':
            for chrom in chroms:
                cmd = make_process_cmd(args, config, chrom)
                print(cmd)
                subprocess.call(cmd, shell=True)
        else:
            process(args, config)

    elif args.action == 'collect':
        collect(args, config)

    elif args.action == 'full':
        for chrom, _ in config.items('chroms'):
            cmd = make_process_cmd(args, config, chrom)
            print(cmd)
            subprocess.call(cmd, shell=True)

        collect_cmd = make_collect_cmd(args, config)
        print(collect_cmd)
        subprocess.call(collect_cmd, shell=True)

    elif args.action == 'check':
        if not all(map(lambda x: os.path.exists(format_outfile(
                config, args.rep, x, 'clusters')),
                       zip(*config.items('chroms'))[0])):
            print('processing is incomplete')
            sys.exit(1)
        print('processing is complete')


if __name__ == '__main__':
    main()
