import ConfigParser

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

from lib5c.util.bed import get_midpoint, parse_feature_from_string
from lib5c.plotters.matrix import plot_matrix

from hiclite.util.filenames import format_outfile
from hiclite.util.sparse_matrix import SparseContactMatrix as SCM
from hiclite.toolbox import subcommand, argument


@subcommand('plot_apa', 'draw aggregate peak analysis pileup', [
    argument('rep'),
    argument('--base_outdir', default='heatmaps'),
    argument('--config', default='config.cfg'),
    argument('--threshold', default=300000)
])
def main(args):
    print('parsing config')
    config = ConfigParser.RawConfigParser(allow_no_value=True)
    config.optionxform = str
    config.read(args.config)
    res = config.getint('assemble_params', 'resolution')

    # load df
    df = pd.read_csv(format_outfile(
        config, args.rep, 'all', 'loops', ext='tsv'), sep='\t', index_col=0)

    # initialize APA grid
    grid = np.zeros((21, 21), dtype=float)

    # iterate over chroms
    for chrom, _ in config.items('chroms'):
        print chrom

        # non-lazy loading of obs (to dodge ~17k lazy reads)
        obs = SCM.load(format_outfile(config, args.rep, chrom, 'obs')).toarray()

        # list of loop centers to be collected
        loop_centers = []

        # look at all loops, 24 passes better than ~17k lazy reads and one pass
        for i in range(len(df.index)):
            # resolve anchors
            loop_id = df.index[i]
            grange_y, grange_x = map(parse_feature_from_string,
                                     loop_id.split('_'))

            # skip loops not on this chrom
            if grange_y['chrom'] != chrom:
                continue

            # establish interaction distance and filter on it
            mid_y = get_midpoint(grange_y, force_int=True)
            mid_x = get_midpoint(grange_x, force_int=True)
            distance = mid_x - mid_y
            if distance < args.threshold:
                continue

            # get bin coordinates
            bin_y = mid_y / res
            bin_x = mid_x / res

            # check for slice-off-edge scenario
            if bin_y < 11 or bin_x < 11 or len(obs) - bin_y < 11 or \
                    len(obs) - bin_x < 11:
                continue

            # append to list
            loop_centers.append((bin_y, bin_x))

        # add submatrix to grid
        grid = np.nansum(np.stack([obs[y-10:y+11, x-10:x+11]
                                   for y, x in loop_centers] + [grid]), axis=0)

    # compute APA score
    score = grid[10, 10] / np.mean(grid[-6:, :6])

    # plot matrix as grid
    plot_matrix(grid, cmap='Reds')
    plt.title('APA score: %s' % score)
    scale = (res * 10) / 1000
    plt.xticks([0, 20], ['-%skb' % scale, '+%skb' % scale])
    plt.yticks([0, 20], ['-%skb' % scale, '+%skb' % scale])
    plt.savefig('%s/%s_apa.png' % (args.base_outdir, args.rep),
                bbox_inches='tight', dpi=800)


if __name__ == '__main__':
    main()
