import ConfigParser

from ast import literal_eval
import matplotlib.pyplot as plt
import scipy.sparse as sparse
import seaborn as sns
import numpy as np
import subprocess
import os

from lib5c.util.plotting import plotter
from lib5c.util.system import check_outdir
from hiclite.util.filenames import format_outfile
from hiclite.util.sparse_matrix import SparseContactMatrix
from hiclite.toolbox import subcommand, argument

def make_histogram_cmd(args, config,chrom=None):
    if chrom is None:
        chrom = args.chrom
    if config.getboolean('scheduler', 'bsub'):
        prefix = 'bsub -M %s -n %s ' % (
            config.get('scheduler', 'process_memory'),
            config.get('scheduler', 'process_cores'))
    else:
        prefix = ''
    return '%shiclite plot_distribution histogram %s --stage %s --chrom %s --config %s' % \
        (prefix, args.rep, args.stage, chrom, args.config)

def histogram(rep,stage,chrom,config):
    outfile = format_outfile(config,rep,chrom,stage,ext='png').split('.')[0]+'_histogram.png'
    
    check_outdir(outfile)
    
    data = SparseContactMatrix.load(format_outfile(config, rep,
                    chrom, stage)).flatten()

    finite_data = data[np.isfinite(data)]
    nonzero_data = finite_data[np.nonzero(finite_data)]

    params = {config.items('distribution')[i][0]: literal_eval(config.items('distribution')[i][1]) 
                    for i in range(len(config.items('distribution')))}

    plt.title('Histogram of %s %s %s'%(rep,chrom,stage))
    plt.xlabel('log2(counts + 1)')
    plt.ylabel('Probability density')
    plt.hist(np.log2(nonzero_data + 1),**params)
    plt.savefig(outfile,dpi=800)

def make_row_sum_cmd(args, config,chrom=None):
    if chrom is None:
        chrom = args.chrom
    if config.getboolean('scheduler', 'bsub'):
        prefix = 'bsub -M %s -n %s ' % (
            config.get('scheduler', 'process_memory'),
            config.get('scheduler', 'process_cores'))
    else:
        prefix = ''
    return '%shiclite plot_distribution row_sum %s --stage %s --chrom %s --config %s' % \
        (prefix, args.rep, args.stage, chrom, args.config)

def row_sum(rep,stage,chrom,config):
    outfile = format_outfile(config,rep,chrom,stage,ext='png').split('.')[0]+'_row_sums.png'

    check_outdir(outfile)

    data = sparse.load_npz(format_outfile(config, rep,
                    chrom, stage)).toarray()

    symmetrized_data = data + data.T

    np.fill_diagonal(symmetrized_data, np.diag(data))

    row_sums = np.nansum(symmetrized_data,axis=0)

    plt.boxplot(row_sums)

    plt.title('Row Sums of %s %s %s'%(rep,chrom,stage))
    plt.ylabel('Row Sum')
    plt.xlabel(chrom)
    plt.savefig(outfile,dpi=800)

def make_sparsity_cmd(args, config,chrom=None):
    if chrom is None:
        chrom = args.chrom
    if config.getboolean('scheduler', 'bsub'):
        prefix = 'bsub -M %s -n %s ' % (
            config.get('scheduler', 'process_memory'),
            config.get('scheduler', 'process_cores'))
    else:
        prefix = ''
    return '%shiclite plot_distribution sparsity %s --stage %s --chrom %s --config %s' % \
        (prefix, args.rep, args.stage, chrom, args.config)


def sparsity(rep,stage,chrom,config):
    outfile = format_outfile(config,rep,chrom,stage,ext='png').split('.')[0]+'_sparsity.png'

    check_outdir(outfile)

    data = SparseContactMatrix.load(format_outfile(config, rep,
                    chrom, stage),max_range=config.getint('obs_params','max_range'))
    data.symmetrize()

    k=config.getint('filter_params', 'sparse_range')

    us_rows = np.where((data.offsets >= -k) & (data.offsets < 0))[0]
    ds_rows = np.where((data.offsets <= k) & (data.offsets > 0))[0]

    nonzero_counts_us = np.sum((data.data[us_rows, :] > 0), axis=0)
    nonzero_counts_ds = np.sum((data.data[ds_rows, :] > 0), axis=0)

    bp_dict = plt.boxplot(nonzero_counts_us)

    for line in bp_dict['medians']:
        x, y = line.get_xydata()[1]
        plt.text(x, y, '%i' % y,verticalalignment='center')

    for line in bp_dict['boxes']:
        print line.get_xydata()
        x, y = line.get_xydata()[1]
        plt.text(x,y, '%i' % y, verticalalignment='center',horizontalalignment='left')
        x, y = line.get_xydata()[2]
        plt.text(x,y, '%i' % y, verticalalignment='center',horizontalalignment='left') 

    for line in bp_dict['caps']:
        x, y = line.get_xydata()[1]
        plt.text(x,y, '%i' % y,verticalalignment='center')

    plt.title('Row Sparsity of %s %s %s within %s bins of the diagonal'%(rep,chrom,stage,k))
    plt.ylabel('Row Sparsity')
    plt.xlabel(chrom)
    plt.savefig(outfile,dpi=800)

@subcommand('plot_distribution', 'plot distribution for ' \
                'quality checking (QC) purposes', [
    argument('action', choices=['histogram','row_sum', 'sparsity']),
    argument('rep'),
    argument('--stage', choices=['input','raw','filtered','balanced',\
                                'obs','exp','pvalues']),
    argument('--chrom', default='all'),
    argument('--config', default='config.cfg')
])
def main(args):

    print('parsing config')
    config = ConfigParser.RawConfigParser(allow_no_value=True)
    config.optionxform = str
    config.read(args.config)
    
    if args.action == 'histogram':
        if args.chrom == 'all':
            for chrom, _ in config.items('chroms'):
                cmd = make_histogram_cmd(args, config, chrom)
                print(cmd)
                subprocess.call(cmd,shell=True)
        else:
            histogram(args.rep,args.stage,args.chrom, config)
    if args.action == 'row_sum':
        if args.chrom == 'all':
            for chrom, _ in config.items('chroms'):
                cmd = make_row_sum_cmd(args, config, chrom)
                print(cmd)
                subprocess.call(cmd,shell=True)
        else:
            row_sum(args.rep,args.stage,args.chrom, config)
    if args.action == 'sparsity':
        if args.chrom == 'all':
            for chrom, _ in config.items('chroms'):
                cmd = make_sparsity_cmd(args, config, chrom)
                print(cmd)
                subprocess.call(cmd,shell=True)
        else:
            sparsity(args.rep,args.stage,args.chrom, config)

    
if __name__ == '__main__':
    main()
