import os
import ConfigParser

import numpy as np
import pandas as pd
import scipy.stats as stats

from lib5c.plotters.heatmap import plot_heatmap
from lib5c.plotters.colormaps import get_colormap
from lib5c.util.bed import parse_feature_from_string, get_midpoint, \
    check_intersect

from hiclite.util.filenames import format_outfile
from hiclite.util.sparse_matrix import SparseContactMatrix as SCM
from hiclite.toolbox import subcommand, argument


def pad_midpoint(chrom, coord, size, res, chrom_size):
    return {'chrom': chrom,
            'start': max((int(coord - 0.5*size*res)/res), 0)*res,
            'end'  : min((int(coord + 0.5*size*res)/res), chrom_size)*res}


def loop_id_to_cluster(loop_id, offset_x, offset_y, resolution):
    y_anchor, x_anchor = map(parse_feature_from_string, loop_id.split('_'))
    return [{'x': i - offset_x, 'y': j - offset_y}
            for i in range(x_anchor['start']/resolution,
                           x_anchor['end']/resolution)
            for j in range(y_anchor['start']/resolution,
                           y_anchor['end']/resolution)]


def check_loop_in_window(loop_id, grange_x, grange_y):
    y_anchor, x_anchor = map(parse_feature_from_string, loop_id.split('_'))
    if check_intersect(x_anchor, grange_x) \
            and check_intersect(y_anchor, grange_y):
        return True
    return False


@subcommand('zoom', 'plot heatmaps of specific zoom windows', [
    argument('x'),
    argument('-y'),
    argument('rep'),
    argument('--stage', choices=['input', 'raw', 'filtered', 'balanced', 'obs',
                                 'deconv_obs', 'exp', 'deconv_exp', 'var',
                                 'pvalues', 'qvalues', 'obs_over_exp',
                                 'interaction_score', 'qis', None]),
    argument('--diag', choices=['on', 'off', None]),
    argument('--tracks', choices=['on', 'off', 'default', None]),
    argument('--nogenes', action='store_true'),
    argument('--size', type=int),
    argument('--vmin', type=float),
    argument('--vmax', type=float),
    argument('--obs_cmap'),
    argument('--cbar', choices=['on', 'off', None]),
    argument('--clusters', choices=['on', 'off', None]),
    argument('--loopfile'),
    argument('--column'),
    argument('--threshold', type=float),
    argument('--outdir'),
    argument('--config', default='config.cfg'),
    argument('--name')
])
def main(args):
    print('parsing config')
    config = ConfigParser.RawConfigParser(allow_no_value=True)
    config.optionxform = str
    config.read(args.config)
    res = config.getint('assemble_params', 'resolution')
    max_range = config.getint('obs_params', 'max_range')

    # resolve args
    stage = config.get('heatmap', 'stage') if args.stage is None else args.stage
    diag = config.get('heatmap', 'diag') if args.diag is None else args.diag
    tracks = config.get('heatmap', 'tracks') if args.tracks is None \
        else args.tracks
    nogenes = True if args.nogenes else \
        config.getboolean('heatmap', 'nogenes') \
        if config.has_option('heatmap', 'nogenes') else False
    size = config.getint('heatmap', 'size') if args.size is None else args.size
    vmin = config.getfloat('heatmap', 'vmin') if args.vmin is None \
        else args.vmin
    vmax = config.getfloat('heatmap', 'vmax') if args.vmax is None \
        else args.vmax
    obs_cmap = config.get('heatmap', 'obs_cmap') if args.obs_cmap is None \
        else args.obs_cmap
    cbar = config.getboolean('heatmap', 'cbar') if args.cbar is None \
        else True if args.cbar == 'on' else False
    clusters = config.getboolean('heatmap', 'clusters') \
        if args.clusters is None else True if args.clusters == 'on' else False
    column = config.get('heatmap', 'column') if args.column is None \
        else args.column
    threshold = config.getfloat('heatmap', 'threshold') \
        if args.threshold is None else args.threshold
    outdir = config.get('heatmap', 'outdir') if args.outdir is None \
        else args.outdir

    # parse id
    x_anchor = parse_feature_from_string(args.x)
    y_anchor = parse_feature_from_string(args.y) if args.y is not None \
        else x_anchor
    chrom = x_anchor['chrom']

    # resolve chrom-dependent args
    loopfile = args.loopfile if args.loopfile is not None else \
        config.get('heatmap', 'loopfile') if \
        config.has_option('heatmap', 'loopfile') else \
        format_outfile(config, args.rep, chrom, 'loops', ext='tsv') if \
        config.has_option('steps', 'clusters') else ''

    # compute midpoints
    x_midpoint = get_midpoint(x_anchor)
    y_midpoint = get_midpoint(y_anchor)
    if diag == 'on':
        midpoint = (x_midpoint + y_midpoint) / 2
        x_midpoint = midpoint
        y_midpoint = midpoint

    # function to compute matrix slices given chromosome size in bins
    def get_slices(chrom_size):
        if size:
            # compute zoom window granges
            g_x = pad_midpoint(chrom, x_midpoint, size, res, chrom_size)
            g_y = pad_midpoint(chrom, y_midpoint, size, res, chrom_size)
        else:
            g_x = x_anchor
            g_y = y_anchor

        # compute slices
        s_x = slice(g_x['start']/res, g_x['end']/res)
        s_y = slice(g_y['start']/res, g_y['end']/res)

        return g_x, g_y, s_x, s_y

    deconv = False
    bias = None
    # obtain matrix slice appropriate for stage
    if stage in ['input', 'raw', 'filtered', 'balanced', 'obs', 'deconv_obs',
                 'exp', 'deconv_exp', 'var', 'pvalues']:
        if 'deconv' in stage:
            deconv = True
        if deconv:
            deconv = True
            stage = stage.replace('deconv_', '')
            bias = np.loadtxt(format_outfile(
                config, args.rep, chrom, 'bias'))
        scm = SCM.load(format_outfile(
            config, args.rep, chrom, stage), max_range=max_range)
        if deconv:
            scm = scm.deconvolute(bias)
        scm.symmetrize()
        grange_x, grange_y, slice_x, slice_y = get_slices(scm.shape[0])
        matrix = scm[slice_y, slice_x]
        cmap = get_colormap(obs_cmap, set_bad='white')
        if stage in ['pvalues', 'qvalues']:
            matrix = stats.norm.isf(matrix)
            cmap = 'seismic'
    elif stage == 'obs_over_exp':
        obs = SCM.load(format_outfile(
            config, args.rep, chrom, 'obs'), max_range=max_range)
        obs.symmetrize()
        exp = SCM.load(format_outfile(
            config, args.rep, chrom, 'exp'), max_range=max_range)
        exp.symmetrize()
        grange_x, grange_y, slice_x, slice_y = get_slices(obs.shape[0])
        matrix = np.log2(obs[slice_y, slice_x] / exp[slice_y, slice_x])
        cmap = get_colormap('obs_over_exp')
    elif stage in ['interaction_score', 'qis']:
        outfile_stage = 'qvalues' if stage == 'qis' else 'pvalues'
        pvalues = SCM.load(format_outfile(
            config, args.rep, chrom, outfile_stage), max_range=max_range)
        pvalues.symmetrize()
        grange_x, grange_y, slice_x, slice_y = get_slices(pvalues.shape[0])
        matrix = -10 * np.log2(pvalues[slice_y, slice_x])
        cmap = get_colormap('is')
    else:
        raise ValueError('invalid --stage option')

    if clusters and 'clusters' in zip(*config.items('steps'))[0] and \
            os.path.exists(loopfile):
        # parse loops file
        df = pd.read_csv(loopfile, sep='\t', index_col=0)

        # force indexing on loop_id
        if 'loop_id' in df.columns:
            df.set_index('loop_id', inplace=True)

        # construct clusters
        if column in df.columns:
            sig_clusters = sum(
                [loop_id_to_cluster(loop_id, slice_x.start, slice_y.start, res)
                 for loop_id in df[df[column] < threshold].index
                 if check_loop_in_window(loop_id, grange_x, grange_y)], [])
            insig_clusters = sum(
                [loop_id_to_cluster(loop_id, slice_x.start, slice_y.start, res)
                 for loop_id in df[df[column] >= threshold].index
                 if check_loop_in_window(loop_id, grange_x, grange_y)], [])
            clusters = [sig_clusters, insig_clusters]
            cluster_colors = ['#40ff00', 'purple']
        else:
            sig_clusters = sum(
                [loop_id_to_cluster(loop_id, slice_x.start, slice_y.start, res)
                 for loop_id in df.index
                 if check_loop_in_window(loop_id, grange_x, grange_y)], [])
            clusters = [sig_clusters]
            cluster_colors = ['#40ff00']
    else:
        clusters = None
        cluster_colors = None

    # prepare gene and chipseq tracks
    include_tracks = diag == 'off'
    if tracks != 'default':
        include_tracks = tracks == 'on'
    if include_tracks:
        genes = config.get('matrix', 'genome') if not nogenes else None
        if config.items('tracks'):
            tracks = [track.replace('%r', args.rep)
                      for track in zip(*config.items('tracks'))[1]]
            track_labels = zip(*config.items('tracks'))[0]
        else:
            tracks = None
            track_labels = None
    else:
        genes = None
        tracks = None
        track_labels = None

    print('plotting')
    plot_heatmap(
        matrix=matrix,
        grange_x=grange_x,
        grange_y=grange_y,
        genes=genes,
        colorscale=(vmin, vmax),
        colormap=cmap,
        colorbar=cbar,
        chipseq_tracks=tracks,
        chipseq_track_labels=track_labels,
        clusters=clusters,
        cluster_colors=cluster_colors,
        outfile='%s/%s-%s-%s_%s-%s-%s_%s_%s%s_%s.png' %
                (outdir,
                 grange_x['chrom'], grange_x['start'], grange_x['end'],
                 grange_y['chrom'], grange_y['start'], grange_y['end'],
                 diag, '%s_' % args.name if args.name else '', args.rep,
                 'deconv_%s' % stage if deconv else stage)
    )


if __name__ == '__main__':
    main()
