import numpy as np
import pandas as pd
import scipy.sparse as sparse

import pyBigWig


def make_mappability_table(restriction_fragment_bedfile, mappability_bigwig,
                           chrom, window_size=500):
    bw = pyBigWig.open(mappability_bigwig)
    df = pd.read_csv(restriction_fragment_bedfile, sep='\t', header=None,
                     usecols=[0, 2], names=['chrom', 'coord'])
    list_of_dict = []
    for row in df[df['chrom'] == chrom].itertuples():
        coord = getattr(row, 'coord')
        list_of_dict.append({
            'coord': coord,
            'side': '-',
            'map': bw.stats(chrom, coord - window_size, coord, exact=True)[0]
        })
        if coord == bw.chroms(chrom):
            continue
        list_of_dict.append({
            'coord': coord,
            'side': '+',
            'map': bw.stats(chrom, coord,
                            min(coord + window_size, bw.chroms(chrom)),
                            exact=True)[0]
        })
    return pd.DataFrame(list_of_dict, columns=['coord', 'side', 'map'])


def make_mappability_vector(mappability_table, resolution=10000):
    df = mappability_table
    df['bin'] = (df['coord'] / resolution).astype(int)
    grouped = df.groupby('bin')['map'].max()
    return sparse.coo_matrix(
        (grouped.as_matrix(),
         (np.zeros(len(grouped), dtype=int),
          np.asarray(grouped.index)))).toarray()[0]


def make_bin_mappability_vector(mappability_bigwig, chrom, resolution=10000,
                                window_width=50000):
    window_radius = window_width / 2
    bw = pyBigWig.open(mappability_bigwig)
    chrom_size_bp = bw.chroms(chrom)
    chrom_size_bin = int(np.ceil((chrom_size_bp / float(resolution))))
    vector = np.zeros(chrom_size_bin)
    for i in range(chrom_size_bin):
        vector[i] = bw.stats(
            chrom,
            max(i*resolution - window_radius, 0),
            min((i+1)*resolution + window_radius, chrom_size_bp),
            exact=True
        )[0]
    return vector


def filter_low_mappability_bins(matrix, mappability_vector, threshold=0.5):
    matrix = matrix.copy()
    keep = sparse.diags([mappability_vector > threshold], [0], dtype=int)
    matrix = keep.dot(matrix)
    matrix = matrix.dot(keep)
    return matrix
