import numpy as np
from statsmodels.nonparametric.smoothers_lowess import lowess


def fit_mvr(obs, exp, min_obs=2, min_exp=0.5, lowess_frac=0.1, delta_frac=0.01,
            return_intermediates=False, return_curve=0):
    """
    Fits an MVR by lowess-smoothing the point-by-point deviations between
    observed and expected data.

    Parameters
    ----------
    obs, exp : np.ndarray
        Parallel arrays containing observed and expected values, respectively.
    min_obs, min_exp : float
        Disregard points whose observed or expected values are smaller than
        these numbers. Points that fail the ``min_exp`` threshold but pass the
        ``min_obs`` threshold will be assigned the variance estimate for the
        smallest expected value greater than or equal to ``min_exp``.
    lowess_frac : float
        The fraction of the data to consider when lowess smoothing.
    delta_frac : float
        The lowess ``delta`` parameter will be set to this fraction of the range
        of the observed values.
    return_intermediates : bool
        Pass True to return a tuple of (var, flat_exp, flat_dev) where flat_exp
        and flat_dev are the subsetted and flattened arrays of expected and
        deviation values used to fit the curve.
    return_curve : int
        If ``return_intermediates`` is True, pass an int greater than zero to
        return a tuple of (var, flat_exp, flat_dev, curve_xs, curve_ys) where
        curve_xs and curve_ys are sorted parallel arrays of expected values and
        variance estimates, respectively, whose length is equal to
        ``return_curve``, sampled uniformly along the distribution of the
        expected values.

    Returns
    -------
    np.ndarray
        Parallel to obs and exp, contains the MVR estimated variance values.
    """
    assert obs.shape == exp.shape
    print('computing deviations')
    dev = (obs - exp)**2
    xs = exp
    ys = dev
    print('masking')
    base_mask = np.isfinite(xs) & np.isfinite(ys) & (obs >= min_obs)
    used_mask = base_mask & (exp >= min_exp)
    fill_mask = base_mask & (exp < min_exp)
    xs = xs[used_mask]
    ys = ys[used_mask]
    print('sorting')
    idx = np.argsort(xs)
    xs = xs[idx]
    ys = ys[idx]
    print('lowess smoothing')
    y_hats = np.expm1(
        lowess(np.log1p(ys), np.log1p(xs), frac=lowess_frac, is_sorted=True,
               delta=(np.log1p(xs[-1]) - np.log1p(xs[0])) * delta_frac,
               return_sorted=False))
    var = np.tile(np.nan, obs.shape)
    sorted_y_hats = np.zeros_like(y_hats)
    sorted_y_hats[idx] = y_hats
    var[used_mask] = sorted_y_hats
    var[fill_mask] = y_hats[0]
    assert np.allclose(var[used_mask][idx], y_hats)

    if return_intermediates:
        if return_curve:
            curve_idx = np.floor(np.linspace(0, len(y_hats)-1, return_curve))\
                .astype(int)
            return var, xs, ys, xs[curve_idx], y_hats[curve_idx]
        return var, xs, ys
    return var
