import numpy as np
import scipy.sparse as sparse
from scipy.ndimage.filters import median_filter

from hiclite.util.sparse_matrix import roll_footprint, \
    SparseContactMatrix as SCM
from hiclite.steps.balance import kr_balance


def flag_high_spatial_outliers(array, size=5, fold_threshold=4.0):
    """
    Identifies which elements of an array are high spatial outliers.

    Parameters
    ----------
    array : np.ndarray
        The array to look for outliers in.
    size : int
        The size of the window to look in around each element when deciding if
        it is an outlier. Should be an odd integer.
    fold_threshold : float
        Elements will be flagged as outliers if they are greater than this
        number or greater than this many times the local median (as estimated
        using the window size in ``size``).

    Returns
    -------
    np.ndarray
        A matrix of the same size and shape as the input matrix, with 1's at
        positions flagged as high spatial outliers and 0's everywhere else.
    """
    median_filtered_array = median_filter(
        array, footprint=roll_footprint(np.ones((size, size), dtype=bool).T))
    flagged_indices = np.zeros_like(array)
    flagged_indices[
        (array > np.maximum(fold_threshold,
                            fold_threshold * median_filtered_array))] = 1
    return flagged_indices


def remove_high_spatial_outliers(matrix, max_range=1000, size=5,
                                 fold_threshold=4.0):
    """
    Convenience function for removing high spatial outliers from sparse
    matrices.

    Operates in-place if the input is CSR.

    Parameters
    ----------
    matrix : scipy.sparse.spmatrix
        The matrix to remove outliers from. Will be converted to CSR.
    max_range : int
        The number of off-diagonals to load when constructing the
        SparseContactMatrix (SCM). Default value corresponds to 3Mb at 4kb
        resolution.
    size : int
        The size of the window to look in around each element when deciding if
        it is an outlier. Should be an odd integer.
    fold_threshold : float
        Elements will be flagged as outliers if they are greater than this
        number or greater than this many times the local median (as estimated
        using the window size in ``size``).

    Returns
    -------
    scipy.sparse.spmatrix
        The input matrix with all spatial outliers overwritten.
    """
    matrix = matrix.tocsr()
    normalized, _, _ = kr_balance(matrix)
    normalized_scm = SCM(normalized, max_range=max_range, upper=True)
    flagged_indices = SCM.apply(flag_high_spatial_outliers,
                                normalized_scm, size=size,
                                fold_threshold=fold_threshold)
    flagged_indices_csr = sparse.csr_matrix(flagged_indices)
    matrix[flagged_indices_csr == 1] = np.nan
    return matrix
