import numpy as np
import scipy.stats as stats


def call_pvalues(frozen_dist, obs, tail='right'):
    """
    Calls left- or right-tail p-values given a frozen distribution and an
    observed value.

    Parameters
    ----------
    frozen_dist : scipy.stats.rv_frozen
        The frozen distribution.
    obs : np.ndarray
        The observed values.
    tail : {'left', 'right'}
        The tail to get p-values for.

    Returns
    -------
    np.ndarray
        The p-values.
    """
    if tail == 'right':
        if hasattr(frozen_dist, 'pmf'):
            obs = np.floor(obs)  # floor here, otherwise pmf will be 0
            return frozen_dist.sf(obs) + frozen_dist.pmf(obs)
        else:
            return frozen_dist.sf(obs)
    elif tail == 'left':
        return frozen_dist.cdf(obs)
    else:
        raise ValueError('invalid tail %s; must be right or left' % tail)


def call_two_tail_poisson_pvalue(obs, exp):
    """
    Calls a two-tailed Poisson p-value given an observed and expected value.

    Based on the logic described in https://stats.stackexchange.com/a/140517

    Parameters
    ----------
    obs, exp : np.ndarray
        The observed and expected values, respectively.

    Returns
    -------
    np.ndarray
        The two-tailed p-values.

    Examples
    --------
    >>> import scipy.stats as stats
    >>> from hiclite.util.pvalues import call_two_tail_poisson_pvalue, \
        call_pvalues
    >>> exp = 10
    >>> dist = stats.poisson(10)
    >>> obs = 15  # right tail example
    >>> right_tail = call_pvalues(dist, obs, 'right')
    >>> call_pvalues(dist, 5, 'left') <= right_tail  # smallest not exceeding
    True
    >>> call_pvalues(dist, 6, 'left') <= right_tail  # next bigger exceeds
    False
    >>> call_two_tail_poisson_pvalue(obs, exp) == \
        right_tail + call_pvalues(dist, 5, 'left')
    True
    >>> obs = 5  # left tail example
    >>> left_tail = call_pvalues(dist, obs, 'left')
    >>> call_pvalues(dist, 16, 'right') <= left_tail  # smallest not exceeding
    True
    >>> call_pvalues(dist, 15, 'right') <= left_tail  # next bigger exceeds
    False
    >>> call_two_tail_poisson_pvalue(obs, exp) == \
        left_tail + call_pvalues(dist, 16, 'right')
    True
    """
    # coerce to array
    obs = np.asarray(obs)
    exp = np.asarray(exp)

    # identify finite positions and filter
    finite_idx = np.isfinite(obs) & np.isfinite(exp)
    obs = obs[finite_idx]
    exp = exp[finite_idx]

    # make sure obs is floored
    obs = np.floor(obs)

    # parametrize a poisson distribution with the given expected value
    dist = stats.poisson(exp)

    # compute right tail and left tail p-values
    right_tail = call_pvalues(dist, obs, 'right')
    left_tail = call_pvalues(dist, obs, 'left')

    # compute "opposite tail" p-values for right tail
    ort_obs = dist.ppf(right_tail) - 1
    ort_obs_idx = np.isfinite(ort_obs)
    opposite_right_tail = np.full_like(ort_obs_idx, np.nan, dtype=float)
    opposite_right_tail[ort_obs_idx] = call_pvalues(
        stats.poisson(exp[ort_obs_idx]), ort_obs[ort_obs_idx], 'left')

    # same for left tail
    olt_eps_idx = left_tail >= 2 * np.finfo(float).eps
    olt_obs = stats.poisson(exp[olt_eps_idx]).isf(left_tail[olt_eps_idx]) + 1
    opposite_left_tail = np.full_like(olt_eps_idx, np.nan, dtype=float)
    opposite_left_tail[olt_eps_idx] = call_pvalues(
        stats.poisson(exp[olt_eps_idx]), olt_obs, 'right')
    opposite_left_tail[~olt_eps_idx] = 2 * left_tail[~olt_eps_idx]

    # prepare, fill in, and return results
    result = np.full_like(finite_idx, np.nan, dtype=float)
    result[finite_idx] = np.where(
        right_tail < left_tail,
        right_tail + opposite_right_tail,
        left_tail + opposite_left_tail
    )

    return result


def approximate_two_tail_poisson_pvalue(obs, exp):
    """
    Computes a conservative approximation of a two-tailed Poisson p-value given
    an observed and expected value.

    Based on the note at the end of https://stats.stackexchange.com/a/140517
    that ``2*t`` should still be an upper bound on the p-value.

    Parameters
    ----------
    obs, exp : np.ndarray
        The observed and expected values, respectively.

    Returns
    -------
    np.ndarray
        The two-tailed p-values.
    """
    # coerce to array
    obs = np.asarray(obs)
    exp = np.asarray(exp)

    # identify finite positions and filter
    finite_idx = np.isfinite(obs) & np.isfinite(exp)
    obs = obs[finite_idx]
    exp = exp[finite_idx]

    # make sure obs is floored
    obs = np.floor(obs)

    # parametrize a poisson distribution with the given expected value
    dist = stats.poisson(exp)

    # compute right tail and left tail p-values
    right_tail = call_pvalues(dist, obs, 'right')
    left_tail = call_pvalues(dist, obs, 'left')

    # prepare, fill in, and return results
    result = np.full_like(finite_idx, np.nan, dtype=float)
    result[finite_idx] = 2 * np.minimum(right_tail, left_tail)
    return result
